""" Collection of functions implementing Gillespie's SSA algorithm and
    conversion to Fokker-Planck and stochastic differential equation
    approximations.
"""


def ssaode(stoichmat,propensfun,rhs,init,tmax,tmin=0,integrator='vode',method='bdf'):
  """ ssaode() - Hybrid SSA-ODE simulation
      uses Gillespie's SSA algorithm in direct method to simulate
      the chemical master equation for species s and an ODE for states x (using
      scipy.integrate.ode)

      rhs: right hand side of ODE like in scipy.integrate.odeint, rhs(t,x,s)
          compatible with scipy.integrate.ode
      propensfun: vector valued propensity function dependent on time, species and 
          states propensfun(t,s,x), describing the likelihood that reaction r_i 
	  happens in the next infinitessimal time step.
      stoichmat: (m x n) stoichiometry matrix with n reactions of m species
      init: initialisation vector, where 
            init[:m] are int values for the species and
            init[m:] are float values for the ode system

      TODO: - do we need a separate stoichiometry matrix, stoimat
            - faster implementation (check append in deque)
	    - partial propensity algorithm 2010 allows scaling with
	      species not with reactions. Use the PDM code from ...
  """
  import scipy, sympy, inspect
  from scipy import integrate
  from collections import deque
  #vars= inspect.getargspec(rhs).args # define variabes a the argument of rhs
  #odedim= len(vars)
  assert scipy.sum(stoichmat,axis=0).any(), "Warning columns of the stoichiometry matrix do not sum to zero: Totoal species numbers are not conserved."
  Nspecies, Nreacts= scipy.shape(stoichmat)
  Node= len(init)-Nspecies
  ode= integrate.ode(rhs).set_integrator(integrator, method=method)
  ode.set_initial_value(init[Nspecies:], tmin)
  y=deque(); y.append(scipy.array(init[Nspecies:],float))
  s=deque(); s.append(scipy.array(init[:Nspecies],int))
  t=deque([0])
  ix= scipy.arange(Nreacts,dtype=scipy.int16)
  while ode.successful() and ode.t < tmax:
    cs= scipy.cumsum(propensfun(t[-1],s[-1],y[-1]))
    tau= -scipy.log(scipy.rand()) / cs[-1] # check sumrate()>0 ?
    j=(ix[ cs > (scipy.rand()*cs[-1]) ])[0]
    ode.set_f_params(s[-1])
    ode.integrate(ode.t+tau)
    y.append(ode.y)
    s.append(s[-1]+stoichmat[:,j])
    t.append(ode.t)
  return scipy.array(t),scipy.array(s,dtype=int),scipy.array(y)

def reaction2digraph(species,nmat,rates):
  """ species : n string vector of species names
      nmat    : nxm stoichiometry matrix 
                (http://en.wikipedia.org/wiki/Stoichiometry#Stoichiometry_matrix)
      rates   : m string vector of reaction rate names
  """
  import networkx, scipy
  species= scipy.asarray(species)
  nmat= scipy.asarray(nmat)
  rates= scipy.asarray(rates)
  return networkx.DiGraph([(str(species[k<0].squeeze()), 
    str(species[k>0].squeeze()),{'label':j}) for j,k in zip(rates,nmat.T)])

def digraph2sdesys(G):
  """ G : networkx.DiGraph object containing the Markov state model
          of the ion channel kinetic scheme
  """
  from networkx import incidence_matrix, to_edgelist
  from sympy import S
  from scipy import zeros, dot, diag, asarray
  rates= asarray([S(k['label']) for i,j,k, in to_edgelist(G)])
  species= asarray([S(k) for k in G.nodes()])
  Nmat= incidence_matrix(G,oriented=1)
  fromreacts= zeros(Nmat.shape)
  fromreacts[Nmat<0]=1
  drift= dot(Nmat,dot(diag(rates),fromreacts.T))
  propensities= dot(species,fromreacts)*rates
  # lumb together unidirectional reactions to bidirectional reactions
  bireact= map(tuple, set(map(frozenset,G.edges())))
  un2bi=[[1 if set(bi)==set(un) else 0 for un in G.edges()] for bi in  bireact]
  # calculate standard deviations as: sqrt(back + forth propensitiy)
  # the abs was used in Orio (2012, Numerical implementation issues)
  stdreacts= map(lambda y: S('sqrt(abs(x))').subs({'x':y}),dot(un2bi,(propensities)))
  reac2species=[[1 if j==k[0] else -1 if j==k[1] else 0 for k in bireact] for j in  map(str,species)]
  smat= dot(reac2species,diag(stdreacts))
  return drift, smat

# def stoich2fpe(N,rates,species):
#   """ Calculates the drift and diffusion matrix """
#   import scipy,sympy
#   from scipy import dot, diag
#   species= map(lambda x: 1.0*sympy.S(x),species)
#   N= scipy.array(N)
#   N2= scipy.zeros(N.shape)
#   N2[N<0]=1
#   drift= scipy.dot(N,scipy.dot(scipy.diag(rates),N2.T))
#   Dspecies= scipy.diag(species)
#   Dspecies= scipy.tile(species,(len(species),1))
#   diffu= dot(N,dot(dot(diag(rates),N.T),diag(species)))
#   #diffu= scipy.dot(drift,Dspecies) + scipy.dot(drift,Dspecies).T
#   return drift,diffu


def digraph2sumOUapprox(channelDiGraph,openstate):
  """ r, v = digraph2sumOUapprox(channelDigraph,openstate)
      Returns
      r : eigenvalues of the infinitesimal Q matrix
      v : list of coeficients
      
      Correlation is: C(t) = const. + p_inf(0) * sum_k v_k * exp(r_k*t)
      v_k = outer(L_k, R_k.T)[openstate,openstate], where L and R are matrices
      with the left and right eigenvectors corresponding to the eigenvalue r_k.
	    
      Method for diffusion approximation of Markov ion channels presented in
      Linaro et al. PLoS Comput Biol (2011).
  """
  import sympy, networkx
  from sympy import S, Matrix, factor, simplify
  from scipy import array, diag
  #G=networkx.DiGraph(channelDiGraph.edges(data='label'))
  G=channelDiGraph
  # formula length is fixed to 1024
  Q=array([[S(i) if i else 0 for i in j] for j in networkx.attr_matrix(G,edge_attr='label',dtype='S1024')[0].tolist()])
  Q=Matrix(Q-diag(Q.sum(1))) # infinitesimal Q matrix
  eigR=Q.eigenvects() # left/right eigensystem
  eigL=Q.T.eigenvects()
  l= [i for i,j,k in eigR] # evals
  r= [i for i,j,k in eigL]
  #print "left/right evals: {} / {}".format(l,r)
  L= Matrix([array(k).flatten().tolist() for i,j,k in eigR]).T
  R= Matrix([array(k).flatten().tolist() for i,j,k in eigL]).T
  # normalise
  D=diag([[simplify(i) for i in j] for j in (L.T*R).tolist()])
  for i,n in enumerate(D):
    R[:,i]/=D[i]
  #print [[simplify(i) for i in j] for j in (L.T*R).tolist()]
  v=[factor(simplify((L[:,j]*R[:,j].T)[openstate,openstate])) for j in range(len(l))]
  return r, v

def digraph2infinitesimalQ(channelDiGraph):
  """ Q = digraph2infinitesimalQ(channelDigraph,openstate)
      Returns
      Q : the infinitesimal Q matrix
      v : list of coeficients
  """
  import sympy, networkx
  from sympy import S, Matrix, factor, simplify
  from scipy import array, diag
  #G=networkx.DiGraph(channelDiGraph.edges(data='label'))
  G=channelDiGraph
  # formula length is fixed to 1024
  Q=array([[S(i) if i else 0 for i in j] for j in networkx.attr_matrix(G,edge_attr='label',dtype='S1024')[0].tolist()])
  Q=Matrix(Q-diag(Q.sum(1))) # invenitesimal Q matrix
  return Q
